{ This is a template generated by Lazarus -> New Project -> FPCUnit Console
  test runner, reworked by Kambi to fit into a reusable unit. }
unit ConsoleTestRunner;

{$mode objfpc}{$H+}

interface

uses
  custapp, classes, sysutils, fpcunit, testreport, testregistry,
  PlainReporter;

Type
  TTestRunner = Class(TCustomApplication)
  private
    FPlainResultsWriter: TPlainResultsWriter;
  protected
    procedure DoRun ; Override;
    procedure doTestRun(aTest: TTest); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  end;

Var
  { Use this to run your tests.
    Just call Application.Initialize, then Application.Run.
    This is created/destroyed in initialization/finalization
    of this unit. }
  Application : TTestRunner;

implementation

Const
  ShortOpts = 'alh';
  Longopts : Array[1..5] of String = (
  'all','list','format:','suite:','help');
  Version = 'Version 0.1';

constructor TTestRunner.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FPlainResultsWriter := TPlainResultsWriter.Create;
end;

destructor TTestRunner.Destroy;
begin
  FPlainResultsWriter.Free;
end;

procedure TTestRunner.doTestRun(aTest: TTest);
var
  testResult: TTestResult;
begin
  testResult := TTestResult.Create;
  try
    testResult.AddListener(FPlainResultsWriter);
    aTest.Run(testResult);
    FPlainResultsWriter.WriteResult(testResult);
  finally
    testResult.Free;
  end;
end;

procedure TTestRunner.DoRun;
var
  I : Integer;
  S : String;
  SuiteFound: boolean;
begin
  S:=CheckOptions(ShortOpts,LongOpts);
  If (S<>'') then
    Writeln(S);
  if HasOption('h', 'help') or (ParamCount = 0) then
  begin
    writeln(Title);
    writeln(Version);
    writeln('Usage: ');
    writeln('-l or --list to show a list of registered tests');
    writeln('default format is xml, add --format=latex to output the list as latex source');
    writeln('-a or --all to run all the tests and show the results in xml format');
    writeln('The results can be redirected to an xml file,');
    writeln('for example: ./testrunner --all > results.xml');
    writeln('use --suite=MyTestSuiteName to run only the tests in a single test suite class');
  end;
  if HasOption('l', 'list') then
  begin
    if HasOption('format') then
    begin
      if GetOptionValue('format') = 'latex' then
        writeln(GetSuiteAsLatex(GetTestRegistry))
      else
        writeln(GetSuiteAsXML(GetTestRegistry));
    end
    else
      writeln(GetSuiteAsXML(GetTestRegistry));
  end;
  if HasOption('a', 'all') then
  begin
    doTestRun(GetTestRegistry)
  end
  else
    if HasOption('suite') then
    begin
      S := '';
      S:=GetOptionValue('suite');
      if S = '' then
        for I := 0 to GetTestRegistry.Tests.count - 1 do
          writeln(GetTestRegistry[i].TestName)
      else
      begin
        SuiteFound := false;
        for I := 0 to GetTestRegistry.Tests.count - 1 do
          if GetTestRegistry[i].TestName = S then
          begin
            SuiteFound := true;
            doTestRun(GetTestRegistry[i]);
          end;
        if not SuiteFound then
          Writeln('Suite name "', S, '" not found');
      end;
    end;
  Terminate;
end;

initialization
  Application:=TTestRunner.Create(Nil);
  Application.Title := 'FPCUnit Console Test Case runner.';
finalization
  FreeAndNil(Application);
end.